/* This file has been put into the public domain by its author.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
   OTHER DEALINGS IN THE SOFTWARE.
*/

/* casinh, casinhf, casinhl - complex arc hyperbolic sine functions

   AUTHOR: Gregory Pietsch

   DESCRIPTION:

   The functionality described on this reference page is aligned with the 
   ISO C standard. Any conflict between the requirements described here and 
   the ISO C standard is unintentional. This volume of POSIX.1-2008 defers 
   to the ISO C standard.

   These functions shall compute the complex arc hyperbolic sine of z, with 
   branch cuts outside the interval [-i, +i] along the imaginary axis.

   RETURN VALUE

   These functions shall return the complex arc hyperbolic sine value, in the 
   range of a strip mathematically unbounded along the real axis and in the 
   interval [-(i * pi/2), +(i * pi/2)] along the imaginary axis.

   Appendix G of the C99 standard gives the following:
   
   casinh(conj(z)) = conj(casinh(z)) and casinh is odd.
 
   casinh(+0 + i0) returns 0 + i0.

   casinh(x + i(Inf)) returns +Inf + i(pi/2) for positive-signed finite x.

   casinh(x + iNaN) returns NaN + iNaN and optionally raises the "invalid"
   floating-point exception, for finite x.

   casinh(+Inf + iy) returns +Inf + i0 for positive-signed finite y.

   casinh(+Inf + i(Inf)) returns +Inf + i(pi/4).

   casinh(+Inf + iNaN) returns +Inf + iNaN.

   casinh(NaN + i0) returns NaN + i0.

   casinh(NaN + iy) returns NaN + iNaN and optionally raises the "invalid"
   floating-point exception, for finite nonzero y.

   casinh(NaN + i(Inf)) returns +/-Inf + iNaN (where the sign of the real 
   part of the result is unspecified).

   casinh(NaN + iNaN) returns NaN + iNaN.

   ERRORS:

   No errors are defined.
*/

#include <complex.h>
#include "xmath.h"

double _Complex (casinh) (double _Complex z)
{
  int cr, ci, sr, si;
  double re, im, t, u;
  double _Complex result;

  re = creal (z);
  im = creal (z);
  cr = _Fpclassify (re);
  ci = _Fpclassify (im);

  switch (cr)
    {
    case FP_NAN:
      if (ci == FP_ZERO || ci == FP_NAN)
	return z;
      else if (ci == FP_INFINITE)
	return CMPLX (im, re);
      else
	{
	  _Matherr (EDOM, FE_INVALID);
	  return CMPLX (re, re);
	}
    case FP_INFINITE:
      sr = _Getsign ((unsigned char *) &re, _Dbl);
      si = _Getsign ((unsigned char *) &im, _Dbl);
      if (sr == 0)
	{
	  if (ci == FP_NAN)
	    return z;
	  else if (ci == FP_INFINITE && si == 0)
	    return CMPLX (re, _M_PI_4L);
	  else if (ci != FP_ZERO)
	    return CMPLX (re, +0.0L);
	}
      break;
    default:
      if (cr == FP_ZERO && ci == FP_ZERO)
	return z;
      if (ci == FP_NAN)
	{
	  _Matherr (EDOM, FE_INVALID);
	  return CMPLX (im, im);
	}
      si = _Getsign ((unsigned char *) &im, _Dbl);
      if (ci == FP_INFINITE && si == 0)
	return CMPLX (_Dbl->_Inf._D, M_PI_2);
      if (ci == FP_ZERO)
	return CMPLX (asinh (re), +0.0);
    }
  t = hypot (re, im + 1);
  u = hypot (re, im - 1);

  return CMPLX (asinh ((t + u) / 2),
		atan2 (im,
		       re + sqrt (re * re +
				  (im + 1) * (im +
					      1))) * (_Signbit (im) ? -1 : 1));
}

float _Complex (casinhf) (float _Complex z)
{
  int cr, ci, sr, si;
  float re, im, t, u;
  float _Complex result;

  re = crealf (z);
  im = crealf (z);
  cr = _Fpclassifyf (re);
  ci = _Fpclassifyf (im);
  switch (cr)
    {
    case FP_NAN:
      if (ci == FP_ZERO || ci == FP_NAN)
	return z;
      else if (ci == FP_INFINITE)
	return CMPLXF (im, re);
      else
	{
	  _Matherr (EDOM, FE_INVALID);
	  return CMPLXF (re, re);
	}
    case FP_INFINITE:
      sr = _Getsign ((unsigned char *) &re, _Flt);
      si = _Getsign ((unsigned char *) &im, _Flt);
      if (sr == 0)
	{
	  if (ci == FP_NAN)
	    return z;
	  else if (ci == FP_INFINITE && si == 0)
	    return CMPLXF (re, _M_PI_4F);
	  else if (ci != FP_ZERO)
	    return CMPLXF (re, +0.0F);
	}
      break;
    default:
      if (cr == FP_ZERO && ci == FP_ZERO)
	return z;
      if (ci == FP_NAN)
	{
	  _Matherr (EDOM, FE_INVALID);
	  return CMPLXF (im, im);
	}
      si = _Getsign ((unsigned char *) &im, _Flt);
      if (ci == FP_INFINITE && si == 0)
	return CMPLXF (_Flt->_Inf._F, _M_PI_2F);
      if (ci == FP_ZERO)
	return CMPLXF (asinhf (re), +0.0F);
    }
  t = hypotf (re, im + 1);
  u = hypotf (re, im - 1);

  return CMPLXF (asinhf ((t + u) / 2),
		 atan2f (im,
			 re + sqrtf (re * re +
				     (im + 1) * (im +
						 1))) *
		 (_Signbitf (im) ? -1 : 1));

}

long double _Complex (casinhl) (long double _Complex z)
{
  int cr, ci, sr, si;
  long double re, im, t, u;
  long double _Complex result;

  re = creall (z);
  im = creall (z);
  cr = _Fpclassifyl (re);
  ci = _Fpclassifyl (im);
  switch (cr)
    {
    case FP_NAN:
      if (ci == FP_ZERO || ci == FP_NAN)
	return z;
      else if (ci == FP_INFINITE)
	return CMPLXL (im, re);
      else
	{
	  _Matherr (EDOM, FE_INVALID);
	  return CMPLXL (re, re);
	}
    case FP_INFINITE:
      sr = _Getsign ((unsigned char *) &re, _Ldbl);
      si = _Getsign ((unsigned char *) &im, _Ldbl);
      if (sr == 0)
	{
	  if (ci == FP_NAN)
	    return z;
	  else if (ci == FP_INFINITE && si == 0)
	    return CMPLXL (re, _M_PI_4L);
	  else if (ci != FP_ZERO)
	    return CMPLXL (re, +0.0L);
	}
      break;
    default:
      if (cr == FP_ZERO && ci == FP_ZERO)
	return z;
      if (ci == FP_NAN)
	{
	  _Matherr (EDOM, FE_INVALID);
	  return CMPLXL (im, im);
	}
      si = _Getsign ((unsigned char *) &im, _Ldbl);
      if (ci == FP_INFINITE && si == 0)
	return CMPLXL (_Ldbl->_Inf._L, _M_PI_2L);
      if (ci == FP_ZERO)
	return CMPLXL (asinhl (re), +0.0);
    }
  t = hypotl (re, im + 1);
  u = hypotl (re, im - 1);

  return CMPLXL (asinhl ((t + u) / 2),
		 atan2l (im,
			 re + sqrt (re * re +
				    (im + 1) * (im +
						1))) *
		 (_Signbitl (im) ? -1 : 1));
}

/* END OF FILE */
