/* This file has been put into the public domain by its author.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
   OTHER DEALINGS IN THE SOFTWARE.
*/

/* csinh, csinhf, csinhl - complex hyperbolic sine functions

   AUTHOR: Gregory Pietsch

   DESCRIPTION:

   The functionality described on this reference page is aligned with the 
   ISO C standard. Any conflict between the requirements described here and 
   the ISO C standard is unintentional. This volume of POSIX.1-2008 defers 
   to the ISO C standard.

   These functions shall compute the complex hyperbolic sine of z.

   RETURN VALUE:

   These functions shall return the complex hyperbolic sine value.

   ERRORS:

   No errors are defined.
   
   Appendix G of the C99 standard lists this behavior:

   - csinh(conj(z)) = conj(csinh(z)) and csinh is odd.

   - csinh(+0 + i0) returns +0 + i0.

   - csinh(+0 + iInf) returns +/-0 + iNaN (where the sign of the real part 
     of the result is unspecified) and raises the "invalid" floating-point 
     exception.

   - csinh(+0 + iNaN) returns +/-0 + iNaN (where the sign of the real part 
     of the result is unspecified).
     
   - csinh(x + iInf) returns NaN + iNaN and raises the "invalid" floating-
     point exception, for positive finite x.

   - csinh(x + iNaN) returns NaN + iNaN and optionally raises the "invalid"
     floating-point exception, for finite nonzero x.
     
   - csinh(+Inf + i0) returns +Inf + i0.
    
   - csinh(+Inf + iy) returns +Inf cis(y), for positive finite y.
    
   - csinh(+Inf + iInf) returns ±Inf + iNaN (where the sign of the real 
     part of the result is unspecified) and raises the "invalid" 
     floating-point exception.

   - csinh(+Inf + iNaN) returns ±Inf + iNaN (where the sign of the real part 
     of the result is unspecified).

   - csinh(NaN + i0) returns NaN + i0.

   - csinh(NaN + iy) returns NaN + iNaN and optionally raises the "invalid"
     floating-point exception, for all nonzero numbers y.

   - csinh(NaN + iNaN) returns NaN + iNaN.
*/

#include <complex.h>
#include "xmath.h"

double _Complex (csinh) (double _Complex z)
{
  double x, y;
  int cx, cy, sx, sy;

  x = creal (z);
  y = cimag (z);
  cx = _Fpclassify (x);
  cy = _Fpclassify (y);

  switch (cx)
    {
    case FP_NAN:
      if (cy == FP_NAN || cy == FP_ZERO)
	return z;
      _Matherr (EDOM, FE_INVALID);
      return CMPLX (x, x);
    case FP_INFINITE:
      sx = _Getsign ((unsigned char *) &x, _Dbl);
      if (sx == 0)
	{
	  switch (cy)
	    {
	    case FP_ZERO:
	      return z;
	    case FP_INFINITE:
	      _Matherr (EDOM, FE_INVALID);	/* FALL THROUGH */
	    case FP_NAN:
	      return (cy == FP_NAN) ? z : CMPLX (x, x);
	    default:
	      return CMPLX (x, x);
	    }
	}
      /* -Infinity is unspecified */
    case FP_ZERO:
      switch (cy)
	{
	case FP_ZERO:
	case FP_NAN:
	  return z;
	case FP_INFINITE:
	  _Matherr (EDOM, FE_INVALID);
	  return CMPLX (x, _Dbl->_Nan._D);
	}
    default:
      switch (cy)
	{
	case FP_ZERO:
	  return CMPLX (sinh (x), y);
	case FP_INFINITE:
	case FP_NAN:
	  _Matherr (EDOM, FE_INVALID);
	  return (cy == FP_NAN) ? CMPLX (y, y) : CMPLX (_Dbl->_Nan._D,
							_Dbl->_Nan._D);
	}
    }
  return CMPLX (sinh (x) * cos (y), cosh (x) * sin (y));
}

float _Complex (csinhf) (float _Complex z)
{
  float x, y;
  int cx, cy, sx, sy;

  x = crealf (z);
  y = cimagf (z);
  cx = _Fpclassifyf (x);
  cy = _Fpclassifyf (y);

  switch (cx)
    {
    case FP_NAN:
      if (cy == FP_NAN || cy == FP_ZERO)
	return z;
      _Matherr (EDOM, FE_INVALID);
      return CMPLXF (x, x);
    case FP_INFINITE:
      sx = _Getsign ((unsigned char *) &x, _Flt);
      if (sx == 0)
	{
	  switch (cy)
	    {
	    case FP_ZERO:
	      return z;
	    case FP_INFINITE:
	      _Matherr (EDOM, FE_INVALID);	/* FALL THROUGH */
	    case FP_NAN:
	      return (cy == FP_NAN) ? z : CMPLXF (x, x);
	    default:
	      return CMPLXF (x, x);
	    }
	}
      /* -Infinity is unspecified */
    case FP_ZERO:
      switch (cy)
	{
	case FP_ZERO:
	case FP_NAN:
	  return z;
	case FP_INFINITE:
	  _Matherr (EDOM, FE_INVALID);
	  return CMPLXF (x, _Flt->_Nan._F);
	}
    default:
      switch (cy)
	{
	case FP_ZERO:
	  return CMPLX (sinhf (x), y);
	case FP_INFINITE:
	case FP_NAN:
	  _Matherr (EDOM, FE_INVALID);
	  return (cy == FP_NAN) ? CMPLX (y, y) : CMPLX (_Flt->_Nan._F,
							_Flt->_Nan._F);
	}
    }
  return CMPLXF (sinhf (x) * cosf (y), coshf (x) * sinf (y));
}

long double _Complex (csinhl) (long double _Complex z)
{
  long double x, y;
  int cx, cy, sx, sy;

  x = creall (z);
  y = cimagl (z);
  cx = _Fpclassifyl (x);
  cy = _Fpclassifyl (y);

  switch (cx)
    {
    case FP_NAN:
      if (cy == FP_NAN || cy == FP_ZERO)
	return z;
      _Matherr (EDOM, FE_INVALID);
      return CMPLXL (x, x);
    case FP_INFINITE:
      sx = _Getsign ((unsigned char *) &x, _Ldbl);
      if (sx == 0)
	{
	  switch (cy)
	    {
	    case FP_ZERO:
	      return z;
	    case FP_INFINITE:
	      _Matherr (EDOM, FE_INVALID);	/* FALL THROUGH */
	    case FP_NAN:
	      return (cy == FP_NAN) ? z : CMPLXL (x, x);
	    default:
	      return CMPLXL (x, x);
	    }
	}
      /* -Infinity is unspecified */
    case FP_ZERO:
      switch (cy)
	{
	case FP_ZERO:
	case FP_NAN:
	  return z;
	case FP_INFINITE:
	  _Matherr (EDOM, FE_INVALID);
	  return CMPLXL (x, _Ldbl->_Nan._L);
	}
    default:
      switch (cy)
	{
	case FP_ZERO:
	  return CMPLXL (sinhl (x), y);
	case FP_INFINITE:
	case FP_NAN:
	  _Matherr (EDOM, FE_INVALID);
	  return (cy == FP_NAN) ? CMPLXL (y, y) : CMPLXL (_Ldbl->_Nan._L,
							  _Ldbl->_Nan._L);
	}
    }

  return CMPLXL (sinhl (x) * cosl (y), coshl (x) * sinl (y));
}

/* END OF FILE */
