/* This file has been put into the public domain by its author.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
   OTHER DEALINGS IN THE SOFTWARE.
*/

/* ctanh, ctanhf, ctanhl - complex hyperbolic tangent functions

   AUTHOR: Gregory Pietsch

   DESCRIPTION:

   The functionality described on this reference page is aligned with the 
   ISO C standard. Any conflict between the requirements described here and 
   the ISO C standard is unintentional. This volume of POSIX.1-2008 defers 
   to the ISO C standard.

   These functions shall compute the complex hyperbolic tangent of z.

   RETURN VALUE:
  
   These functions shall return the complex hyperbolic tangent value.

   ERRORS:

   No errors are defined.

   Appendix G of the C99 standard lists this behavior:
   
   - ctanh(conj(z)) = conj(ctanh(z)) and ctanh is odd.

   - ctanh(+0 + i0) returns +0 + i0.

   - ctanh(x + iInf) returns NaN + iNaN and raises the "invalid" 
     floating-point exception, for finite x.
     
   - ctanh(x + iNaN) returns NaN + iNaN and optionally raises the "invalid" 
     floating-point exception, for finite x.

   - ctanh(+Inf + iy) returns 1 + i0 sin(2y), for positive-signed finite y.

   - ctanh(+Inf + iInf) returns 1 +/- i0 (where the sign of the imaginary part 
     of the result is unspecified).
     
   - ctanh(+Inf + iNaN) returns 1 +/- i0 (where the sign of the imaginary part
     of the result is unspecified).

   - ctanh(NaN + i0) returns NaN + i0.
   
   - ctanh(NaN + iy) returns NaN + iNaN and optionally raises the "invalid" 
     floating-point exception, for all nonzero numbers y.

   - ctanh(NaN + iNaN) returns NaN + iNaN.

*/

#include <complex.h>
#include "xmath.h"

double _Complex (ctanh) (double _Complex z)
{
  double x = creal (z);
  double y = cimag (z);
  int cx, cy;

  cx = _Fpclassify (x);
  cy = _Fpclassify (y);

  switch (cx)
    {
    case FP_NAN:
      if (cy == FP_ZERO || cy == FP_NAN)
	return z;
      else
	{
	  _Matherr (EDOM, FE_INVALID);
	  return CMPLX (x, x);
	}
    case FP_INFINITE:
      return CMPLX (1.0, 0.0);
    case FP_ZERO:
      if (cy == FP_ZERO)
	return z;
      else
	break;
    default:
      if (cy == FP_NAN || cy == FP_INFINITE)
	return (cy == FP_NAN) ? CMPLX (y, y) : CMPLX (_Dbl->_Nan._D,
						      _Dbl->_Nan._D);
      break;
    }

  double sinh_2x = sinh (2.0 * x);
  double sin_2y = sin (2.0 * y);
  double cosh_2x = cosh (2.0 * x);
  double cos_2y = cos (2.0 * y);
  double real = sinh_2x / cosh_2x;
  double imag = sin_2y / cosh_2x;
  return CMPLX (real, imag);
}

float _Complex (ctanhf) (float _Complex z)
{
  float x = crealf (z);
  float y = cimagf (z);
  int cx, cy;

  cx = _Fpclassifyf (x);
  cy = _Fpclassifyf (y);

  switch (cx)
    {
    case FP_NAN:
      if (cy == FP_ZERO || cy == FP_NAN)
	return z;
      else
	{
	  _Matherr (EDOM, FE_INVALID);
	  return CMPLXF (x, x);
	}
    case FP_INFINITE:
      return CMPLXF (1.0F, 0.0F);
    case FP_ZERO:
      if (cy == FP_ZERO)
	return z;
      else
	break;
    default:
      if (cy == FP_NAN || cy == FP_INFINITE)
	return (cy == FP_NAN) ? CMPLXF (y, y) : CMPLXF (_Flt->_Nan._F,
							_Flt->_Nan._F);
      break;
    }

  float sinh_2x = sinhf (2.0L * x);
  float sin_2y = sinf (2.0L * y);
  float cosh_2x = coshf (2.0L * x);
  float cos_2y = cosf (2.0L * y);
  float real = sinh_2x / cosh_2x;
  float imag = sin_2y / cosh_2x;
  return CMPLXF (real, imag);
}

long double _Complex (ctanhl) (long double _Complex z)
{
  long double x = creall (z);
  long double y = cimagl (z);
  int cx, cy;

  cx = _Fpclassifyl (x);
  cy = _Fpclassifyl (y);

  switch (cx)
    {
    case FP_NAN:
      if (cy == FP_ZERO || cy == FP_NAN)
	return z;
      else
	{
	  _Matherr (EDOM, FE_INVALID);
	  return CMPLXL (x, x);
	}
    case FP_INFINITE:
      return CMPLXL (1.0L, 0.0L);
    case FP_ZERO:
      if (cy == FP_ZERO)
	return z;
      else
	break;
    default:
      if (cy == FP_NAN || cy == FP_INFINITE)
	return (cy == FP_NAN) ? CMPLXL (y, y) : CMPLXL (_Ldbl->_Nan._L,
							_Ldbl->_Nan._L);
      break;
    }

  long double sinh_2x = sinhl (2.0L * x);
  long double sin_2y = sinl (2.0L * y);
  long double cosh_2x = coshl (2.0L * x);
  long double cos_2y = cosl (2.0L * y);
  long double real = sinh_2x / cosh_2x;
  long double imag = sin_2y / cosh_2x;
  return CMPLXL (real, imag);
}

/* END OF FILE */
